import re
import logging
from typing import List, Match

from lsprotocol.types import Position
from pygls.workspace import TextDocument
from bibli_ls.bibli_config import CiteConfig

logger = logging.getLogger(__name__)


def clean_list(input: List[str]):
    """Strip and Clear any empty components from the list"""
    return [k.strip() for k in input if k.strip()]


def find_cites(text: str, cite_config: CiteConfig) -> List[Match[str]] | None:
    trigger = cite_config.trigger
    email_positions = [
        (m.start(), m.end()) for m in re.finditer(r"\b[\w\.-]+@[\w\.-]+\.\w+\b", text)
    ]

    # Pattern for citation keys
    pattern = rf"{trigger}([A-Za-z0-9_-]+)"

    cite_match = []
    for match in re.finditer(pattern, text):
        start_pos = match.start()
        is_email = any(start <= start_pos < end for start, end in email_positions)

        if not is_email:
            cite_match.append(match)
    return cite_match


def citekey_at_position(
    doc: TextDocument, position: Position, cite_config: CiteConfig
) -> str | None:
    line = doc.lines[position.line]
    cite_matches = find_cites(line, cite_config)
    if not cite_matches:
        return None

    for match in cite_matches:
        key = match.group(1)
        if position.character >= match.start() and position.character < match.end():
            logger.debug(f"Returning {key}")
            return key

    return None



from pathlib import Path
from typing import Union

from bibtexparser.library import Library
from bibtexparser.model import Block, Entry
from typing_extensions import List


class BibliLibrary(Library):
    path: Path | None

    def __init__(self, blocks: Union[List[Block], None] = None, path=None):
        super().__init__(blocks)
        self.path = path


class BibliBibDatabase:
    libraries: dict[str, list[BibliLibrary]]

    def __init__(self) -> None:
        self.libraries = {}

    def find_in_libraries(
        self, key: str
    ) -> tuple[Entry, BibliLibrary] | tuple[None, None]:
        for libs in self.libraries.values():
            for lib in libs:
                if lib.entries_dict.__contains__(key):
                    return lib.entries_dict[key], lib
        return None, None



from abc import abstractmethod
from bibtexparser.middlewares.names import List
from lsprotocol.types import (
    WorkDoneProgressBegin,
    WorkDoneProgressEnd,
    WorkDoneProgressReport,
)
from pygls.lsp.server import LanguageServer
from pygls.progress import Progress

from bibli_ls.bibli_config import BackendConfig
from bibli_ls.database import BibliLibrary
from bibli_ls.utils import show_message


class BibliBackend:
    backend_type: str
    _config: BackendConfig
    _ls: LanguageServer
    _name: str

    def __init__(self, name: str, config: BackendConfig, ls: LanguageServer):
        self._config = config
        self._ls = ls
        self._name = name

    @abstractmethod
    def get_libraries(self) -> List[BibliLibrary]:
        """Return the libraries based on backend config"""
        pass

    def load_progress_begin(self, location):
        self._progress = Progress(self._ls.protocol)
        self._progress.create(self._name)

        self._progress.begin(
            self._name,
            WorkDoneProgressBegin(
                title=f"Retriving backend type `{self._config.backend_type}` from `{location}`",
                # message="libraries loaded",
            ),
        )

    def load_progress_update(self, msg, loaded, total):
        self._progress.report(
            self._name,
            WorkDoneProgressReport(
                message=msg,
                percentage=int(loaded * 100 / total),
            ),
        )

    def load_progress_done(self, loaded, location):
        show_message(self._ls, f"Loaded {loaded} entries from `{location}`")
        self._progress.end(
            self._name,
            WorkDoneProgressEnd(message="Done"),
        )



import os
from pathlib import Path
from bibtexparser.middlewares.latex_encoding import logging
from bibtexparser.writer import Library
from lsprotocol.types import MessageType
from pygls.lsp.server import LanguageServer
from pyzotero.zotero import bibtexparser
from bibli_ls.backends.backend import BibliBackend
from bibli_ls.bibli_config import BackendConfig
from bibli_ls.database import BibliLibrary

logger = logging.getLogger(__name__)


class BibfileBackend(BibliBackend):
    def __init__(self, name: str, config: BackendConfig, ls: LanguageServer) -> None:
        super().__init__(name, config, ls)

        """TODO: Get all bibtex files found if config is not given."""
        if config.bibfiles == []:
            logger.warning("No bibfile found.", MessageType.Warning)

    def get_libraries(self):
        libraries = []
        total_files = len(self._config.bibfiles)
        loaded_files = 0
        total_entries = 0
        self.load_progress_begin(f"{self._config.bibfiles}")
        for bibfile_path in self._config.bibfiles:
            if not os.path.isabs(bibfile_path) and self._ls.workspace.root_path:
                bibfile_path = os.path.join(self._ls.workspace.root_path, bibfile_path)

            with open(bibfile_path, "r") as bibtex_file:
                library: Library = bibtexparser.parse_string(bibtex_file.read())
                total_entries += len(library.entries)

                libraries.append(
                    BibliLibrary(
                        library.blocks,
                        Path(bibfile_path),
                    )
                )
                self.load_progress_update(bibfile_path, loaded_files, total_files)
            loaded_files += 1
        self.load_progress_done(total_entries, f"{self._config.bibfiles}")
        return libraries



import logging
import os
from pathlib import Path
from typing import Any, Optional

from lsprotocol import types
from pygls.lsp.server import LanguageServer
from pygls.protocol.language_server import LanguageServerProtocol, lsp_method
from pygls.workspace.text_document import TextDocument

from bibli_ls.backends.bibtex_backend import BibfileBackend
from bibli_ls.backends.zotero_backend import ZoteroBackend

from . import __version__
from .bibli_config import BibliTomlConfig
from .database import BibliBibDatabase
from .utils import (
    build_doc_string,
    get_cite_uri,
    show_message,
)
from .parse import citekey_at_position, find_cites

logger = logging.getLogger(__name__)

CONFIG = BibliTomlConfig()
CONFIG_FILE: Path
DATABASE = BibliBibDatabase()


def try_load_configs_file(ls: LanguageServer, root_path=None, config_file=None):
    """Load config file located at the root of the project.
    Use default config if not found.
    """
    import tosholi

    if not config_file:
        if root_path:
            config_file = Path(os.path.join(root_path, ".bibli.toml"))

    if not config_file:
        return

    try:
        global CONFIG, CONFIG_FILE
        f = open(config_file, "rb")

        CONFIG = tosholi.load(BibliTomlConfig, f)  # type: ignore
        CONFIG_FILE = config_file
        show_message(ls, f"Loaded configs from `{config_file}`")
    except NameError as e:
        show_message(
            ls,
            f"Failed to parse config file `{config_file}` error {e}\n",
            types.MessageType.Error,
        )
    except FileNotFoundError:
        show_message(ls, "No config file found, using default settings\n")

    if not CONFIG.sanitize():
        logger.error("Invalid config")


def load_libraries(ls: LanguageServer, use_cached: bool = True):
    global DATABASE, CONFIG
    for k, v in CONFIG.backends.items():
        show_message(
            ls,
            f"Processing backend `{k}` type `{v.backend_type}`",
        )
        if v.backend_type == "zotero_api":
            if not use_cached:
                DATABASE.libraries[k] = ZoteroBackend(k, v, ls).get_libraries()
            else:
                DATABASE.libraries[k] = ZoteroBackend(k, v, ls).get_libraries_cached()

        elif v.backend_type == "bibfile":
            DATABASE.libraries[k] = BibfileBackend(k, v, ls).get_libraries()
        else:
            show_message(
                ls,
                f"Unknown backend type {v.backend_type} ",
                types.MessageType.Error,
            )


class BibliLanguageServerProtocol(LanguageServerProtocol):
    """Override some built-in functions."""

    @lsp_method(types.INITIALIZE)
    def lsp_initialize(self, params: types.InitializeParams) -> types.InitializeResult:
        """Initialize LSP"""

        initialize_result: types.InitializeResult = super().lsp_initialize(params)

        if params.root_path:
            try_load_configs_file(self._server, root_path=params.root_path)

        # Load libraries
        load_libraries(self._server)

        # Register additional trigger characters
        completion_provider = initialize_result.capabilities.completion_provider
        if completion_provider:
            completion_provider.trigger_characters = [
                CONFIG.cite.trigger,
                CONFIG.cite.prefix,
                CONFIG.cite.separator,
            ]

        return initialize_result


class BibliLanguageServer(LanguageServer):
    """Bibli language server.

    :attr initialization_options: initialized in lsp_initialize from the
        protocol_cls.
    """

    # initialization_options: InitializationOptions

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.index = {}
        self.diagnostics = {}
        self.completion_cache = []

        super().__init__(*args, **kwargs)

    def rebuild_completion_items(
        self,
    ):
        processed_keys = {}
        self.completion_cache.clear()
        for libraries in DATABASE.libraries.values():
            for lib in libraries:
                for k, entry in lib.entries_dict.items():
                    key = CONFIG.cite.trigger + k
                    text_edits = []
                    doc_string = build_doc_string(
                        entry, CONFIG.completion.doc_format, str(lib.path)
                    )

                    # Avoid showing duplicated entries
                    if not processed_keys.get(key):
                        processed_keys[key] = True
                        self.completion_cache.append(
                            types.CompletionItem(
                                key,
                                insert_text=key,
                                commit_characters=[
                                    CONFIG.cite.postfix,
                                    CONFIG.cite.separator,
                                ],
                                additional_text_edits=text_edits,
                                kind=types.CompletionItemKind.Reference,
                                documentation=types.MarkupContent(
                                    kind=types.MarkupKind.Markdown,
                                    value=doc_string,
                                ),
                            )
                        )

    def diagnose(self, document: TextDocument):
        global CONFIG
        diagnostics = []

        for idx, line in enumerate(document.lines):
            cite_matches = find_cites(line, CONFIG.cite)
            if not cite_matches:
                continue
            for match in cite_matches:
                key = match.group(1)

                if DATABASE.find_in_libraries(key) != (
                    None,
                    None,
                ):
                    continue

                message = f'Item "{key}" does not exist in library'
                severity = types.DiagnosticSeverity.Warning
                diagnostics.append(
                    types.Diagnostic(
                        message=message,
                        severity=severity,
                        range=types.Range(
                            start=types.Position(line=idx, character=match.start()),
                            end=types.Position(line=idx, character=match.end()),
                        ),
                    )
                )

        self.diagnostics[document.uri] = (document.version, diagnostics)


SERVER = BibliLanguageServer(
    name="bibli-language-server",
    version=__version__,
    protocol_cls=BibliLanguageServerProtocol,
)


@SERVER.thread()
@SERVER.command("library.reload_all")
def reload_all(ls: BibliLanguageServer, *args):
    load_libraries(ls, False)


@SERVER.feature(
    types.TEXT_DOCUMENT_CODE_ACTION,
    types.CodeActionOptions(code_action_kinds=[types.CodeActionKind.Empty]),
)
def code_actions(params: types.CodeActionParams):
    items = []

    items.append(
        types.CodeAction(
            "Bibli: Reload all libraries",
            kind=types.CodeActionKind.Empty,
            command=types.Command("ASDASD", "library.reload_all"),
        )
    )
    return items


@SERVER.feature(types.TEXT_DOCUMENT_DID_SAVE)
def did_save(_: BibliLanguageServer, params: types.DidSaveTextDocumentParams):
    if params.text_document.uri == CONFIG_FILE.as_uri():
        logger.info(f"Config file `{CONFIG_FILE}` modified")


@SERVER.feature(types.TEXT_DOCUMENT_DID_OPEN)
def did_open(ls: BibliLanguageServer, params: types.DidOpenTextDocumentParams):
    """Parse each document when it is opened"""
    doc = ls.workspace.get_text_document(params.text_document.uri)
    ls.diagnose(doc)

    for uri, (version, diagnostics) in ls.diagnostics.items():
        ls.text_document_publish_diagnostics(
            types.PublishDiagnosticsParams(
                uri=uri, version=version, diagnostics=diagnostics
            )
        )


@SERVER.feature(types.TEXT_DOCUMENT_DID_CHANGE)
def did_change(ls: BibliLanguageServer, params: types.DidOpenTextDocumentParams):
    """Parse each document when it is changed"""
    doc = ls.workspace.get_text_document(params.text_document.uri)
    ls.diagnose(doc)

    for uri, (version, diagnostics) in ls.diagnostics.items():
        ls.text_document_publish_diagnostics(
            types.PublishDiagnosticsParams(
                uri=uri, version=version, diagnostics=diagnostics
            )
        )


@SERVER.feature(types.TEXT_DOCUMENT_REFERENCES)
def find_references(ls: BibliLanguageServer, params: types.ReferenceParams):
    """textDocument/references: Find references of an object through simple ripgrep."""

    from ripgrepy import Ripgrepy

    root_path = ls.workspace.root_path
    if not root_path:
        return

    document = ls.workspace.get_text_document(params.text_document.uri)
    cite = citekey_at_position(document, params.position, CONFIG.cite)

    if not cite:
        return

    # Include trigger for better accuracy
    rg = Ripgrepy(CONFIG.cite.trigger + cite, root_path)
    result = rg.with_filename().json().run().as_dict
    references = []

    for res in result:
        if res["type"] == "match":
            # for submatch in res["data"]["submatches"]:
            submatch = res["data"]["submatches"][0]
            file_uri = "file://" + res["data"]["path"]["text"]
            line_no = res["data"]["line_number"]
            references.append(
                types.Location(
                    uri=file_uri,
                    range=types.Range(
                        start=types.Position(
                            line=line_no - 1,
                            character=submatch["start"],
                        ),
                        end=types.Position(
                            line=line_no - 1, character=submatch["end"] - 1
                        ),
                    ),
                )
            )

    # logger.debug(f"RESULTS: {references}")
    return references


@SERVER.feature(types.TEXT_DOCUMENT_DEFINITION)
def goto_definition(ls: BibliLanguageServer, params: types.DefinitionParams):
    """textDocument/definition: Jump to an object's definition."""

    definitions: types.Definition = []

    document = ls.workspace.get_text_document(params.text_document.uri)

    cite = citekey_at_position(document, params.position, CONFIG.cite)
    if not cite:
        return []

    (entry, library) = DATABASE.find_in_libraries(cite)
    if entry and library and library.path is not None:
        from ripgrepy import Ripgrepy

        rg = Ripgrepy(cite, str(library.path))
        result = rg.with_filename().json().run().as_dict
        for res in result:
            if res["type"] == "match":
                submatch = res["data"]["submatches"][0]
                line_no = res["data"]["line_number"]
                definitions.append(
                    types.Location(
                        uri=Path(library.path).as_uri(),
                        range=types.Range(
                            start=types.Position(
                                line=line_no - 1, character=submatch["start"]
                            ),
                            end=types.Position(
                                line=line_no - 1, character=submatch["end"] - 1
                            ),
                        ),
                    )
                )
    # logger.debug(f"Founr definitions: {definitions}")
    return definitions


@SERVER.feature(types.TEXT_DOCUMENT_IMPLEMENTATION)
def goto_implementation(ls: BibliLanguageServer, params: types.DefinitionParams):
    """textDocument/definition: Jump to an object's type definition."""
    document = ls.workspace.get_text_document(params.text_document.uri)

    cite = citekey_at_position(document, params.position, CONFIG.cite)
    if not cite:
        return

    uri = get_cite_uri(DATABASE, cite, CONFIG)
    if uri:
        ls.window_show_document(types.ShowDocumentParams(uri, external=True))
        return types.Location(
            params.text_document.uri, types.Range(params.position, params.position)
        )

    return None


@SERVER.feature(types.TEXT_DOCUMENT_DIAGNOSTIC)
def diagnostic(ls: BibliLanguageServer, params: types.DocumentDiagnosticParams):
    doc = ls.workspace.get_text_document(params.text_document.uri)
    ls.diagnose(doc)

    return types.RelatedFullDocumentDiagnosticReport(ls.diagnostics[doc.uri][1])


@SERVER.feature(types.TEXT_DOCUMENT_HOVER)
def hover(ls: BibliLanguageServer, params: types.HoverParams):
    """textDocument/hover: Display entry metadata."""

    pos = params.position
    document_uri = params.text_document.uri
    document = ls.workspace.get_text_document(document_uri)

    cite = citekey_at_position(document, params.position, CONFIG.cite)
    if not cite:
        return None

    (entry, library) = DATABASE.find_in_libraries(cite)
    if entry and library and library.path:
        hover_text = build_doc_string(entry, CONFIG.hover.doc_format, str(library.path))

        return types.Hover(
            contents=types.MarkupContent(
                kind=types.MarkupKind.Markdown,
                value=hover_text,
            ),
            range=types.Range(
                start=types.Position(line=pos.line, character=0),
                end=types.Position(line=pos.line + 1, character=0),
            ),
        )
    return None


@SERVER.feature(
    types.TEXT_DOCUMENT_COMPLETION,
    types.CompletionOptions(
        # resolve_provider=True,
    ),
)
def completion(
    ls: BibliLanguageServer, params: types.CompletionParams
) -> Optional[types.CompletionList]:
    """textDocument/completion: Returns completion items."""

    document_uri = params.text_document.uri
    document = ls.workspace.get_text_document(document_uri)

    # Heuristics to support cancelled completion with trigger.
    # Even if there is not a trigger, if there is a *kinda* cite at the position,
    # also trigger completion.

    should_complete = False
    char_at_pos = document.lines[params.position.line][params.position.character - 1]
    if (
        char_at_pos == CONFIG.cite.trigger
        or char_at_pos == CONFIG.cite.prefix
        or char_at_pos == CONFIG.cite.separator
    ):
        should_complete |= True

    cite = citekey_at_position(
        document,
        types.Position(params.position.line, params.position.character - 1),
        CONFIG.cite,
    )
    if cite:
        should_complete |= True

    if params.context and (
        params.context.trigger_character == CONFIG.cite.trigger
        or params.context.trigger_character == CONFIG.cite.prefix
        or params.context.trigger_character == CONFIG.cite.separator
    ):
        should_complete |= True

    if not should_complete:
        return None

    if ls.completion_cache == []:
        ls.rebuild_completion_items()

    return types.CompletionList(is_incomplete=False, items=ls.completion_cache)



import re
from typing import List

from bibtexparser.exceptions import ParserStateException, ParsingException
from bibtexparser.model import Entry, Field
from lsprotocol.types import MessageType, Position, ShowMessageParams
from pygls.lsp.server import LanguageServer
from pygls.workspace import TextDocument
import logging
from typing_extensions import assert_type
import requests

from bibli_ls.database import BibliBibDatabase


from .bibli_config import BibliTomlConfig, DocFormatingConfig

logger = logging.getLogger(__name__)


def show_message(
    ls: LanguageServer, msg: str, msg_type: MessageType = MessageType.Info
):
    ls.window_show_message(
        ShowMessageParams(
            msg_type,
            msg,
        )
    )


def get_item_attachments_bbt(cite: str):
    url = "http://localhost:23119/better-bibtex/json-rpc"

    headers = {"Content-Type": "application/json", "Accept": "application/json"}
    payload = {"jsonrpc": "2.0", "method": "item.attachments", "params": [cite]}

    attachments = []
    response = requests.post(url, json=payload, headers=headers)
    data = response.json()
    for attachment in data["result"]:
        attachments.append(attachment["open"])

    return attachments


def get_cite_uri(
    db: BibliBibDatabase, cite: str, config: BibliTomlConfig
) -> str | None:
    (entry, _) = db.find_in_libraries(cite)

    if not entry:
        return None

    uri = None
    match config.view.viewer:
        case "browser":
            if entry.fields_dict.get("url"):
                uri = entry.fields_dict["url"].value
        case "zotero":
            uri = f"zotero://select/items/{cite}"

        case "zotero_bbt":
            # Hacky way to  use better-bibtex JSON RPC support to get the attachment
            # uri. See more:
            # - https://retorque.re/zotero-better-bibtex/exporting/json-rpc/index.html
            # - https://github.com/retorquere/zotero-better-bibtex/issues/1347
            uri = get_item_attachments_bbt(cite)[0]
    return uri

def preprocess_bib_entry(entry: Entry, config: DocFormatingConfig):
    replace_list = ["{{", "}}", "\\vphantom", "\\{", "\\}"]
    for f in entry.fields:
        if isinstance(f.value, str):
            for r in replace_list:
                f.value = f.value.replace(r, "")

            f.value = f.value.replace("\n", " ")
            if len(f.value) > config.character_limit:
                f.value = f.value[: config.character_limit] + "..."

            entry.set_field(Field(f.key, f.value))


def build_doc_string(
    entry: Entry, config: DocFormatingConfig, bibfile: str | None = None
):
    import mdformat

    preprocess_bib_entry(entry, config)

    field_dict = {f.key: f.value for f in entry.fields}

    field_dict["entry_type"] = entry.entry_type
    if bibfile:
        field_dict["bibfile"] = bibfile

    doc_string = ""

    while True:
        try:
            if isinstance(config.header_format, List):
                doc_string += "\n".join(config.header_format).format(**field_dict)
            else:
                assert_type(config.header_format, str)
                doc_string += config.header_format.format(**field_dict)
        except KeyError as e:
            # Unknown key in the header format
            field_dict[e.args[0]] = "Unknown"
            continue
        break

    doc_string += "\n"

    # Entry type no longer needed
    field_dict.pop("entry_type")
    if config.show_fields == []:
        show_field_dict = {k: v for k, v in field_dict.items()}
    else:
        show_field_dict = {
            k: v for k, v in field_dict.items() if config.show_fields.count(k) > 0
        }

    match config.format:
        case "table":
            from py_markdown_table.markdown_table import markdown_table

            table_content = [{"Key": k, "Value": v} for k, v in show_field_dict.items()]
            # NOTE: sometimes table formatting fails with long entries
            table = (
                markdown_table(table_content)
                .set_params(
                    row_sep="markdown",
                    padding_weight="right",
                    multiline={
                        "Key": 20,
                        "Value": 80,
                    },
                    quote=False,
                )
                .get_markdown()
            )
            doc_string += table
        case "list":
            for k, v in show_field_dict.items():
                doc_string += f"- __{k}__: {v}\n"

            # Do one last beautifying for list
            doc_string = mdformat.text(
                doc_string,
                options={"wrap": config.wrap},
            )

    while True:
        try:
            if isinstance(config.footer_format, List):
                doc_string += "\n".join(config.footer_format).format(**field_dict)
            else:
                assert_type(config.footer_format, str)
                doc_string += config.footer_format.format(**field_dict)
        except KeyError as e:
            # Unknown key in the header format
            field_dict[e.args[0]] = "Unknown"
            continue
        break

    return doc_string



import logging
import os
from pathlib import Path

from bibtexparser import bibtexparser
from bibtexparser.middlewares.names import List
from pygls.lsp.server import LanguageServer
from pyzotero.zotero import Zotero

from bibli_ls.backends.backend import BibliBackend
from bibli_ls.backends.bibtex_backend import BibfileBackend
from bibli_ls.bibli_config import BackendConfig
from bibli_ls.database import BibliLibrary
from bibli_ls.utils import show_message

logger = logging.getLogger(__name__)


class ZoteroBackend(BibliBackend):
    _zot: Zotero

    def __init__(self, name: str, config: BackendConfig, ls: LanguageServer):
        super().__init__(name, config, ls)
        if config.library_id == "":
            logger.error("Library ID not specified")
            return

        if config.api_key is None:
            logger.error("API key not specified")
            return

        logger.info(
            f"Initializing zotero API connection library_id `{config.library_id}`, library_type `{config.library_type}`",
        )
        self._zot = Zotero(config.library_id, config.library_type, config.api_key)

    def get_cache_file_path(self) -> str | None:
        if not self._ls.workspace.root_path:
            return None

        filename = f".{self._name}_{self._zot.library_type}_{self._zot.library_id}.bib"
        return os.path.join(self._ls.workspace.root_path, filename)

    def get_libraries_cached(self) -> List[BibliLibrary]:
        cache_file = self.get_cache_file_path()

        if cache_file and os.path.exists(Path(cache_file)):
            show_message(self._ls, f"Loading from cached library `{cache_file}`")
            bibtex_backend = BibfileBackend(
                "cache_file",
                BackendConfig(backend_type="bibfile", bibfiles=[cache_file]),
                self._ls,
            )
            return bibtex_backend.get_libraries()
        else:
            return self.get_libraries()

    def get_libraries(self):
        count = self._zot.count_items()
        loaded = 0
        limit = 100

        show_message(
            self._ls,
            f"Fetching `{count}` items from `{self._zot.library_type}` library `{self._zot.library_id}`",
        )

        self.library = BibliLibrary(path=self.get_cache_file_path())

        self.load_progress_begin(self._zot.library_id)

        for i in range(0, count, limit):
            items = self._zot.items(start=i, limit=limit, content="bibtex")
            logger.debug(items)
            loaded += len(items)
            items_str = "\n".join(items)
            bibtexparser.parse_string(items_str, library=self.library)
            self.load_progress_update(self._zot.library_id, loaded, count)

        self.load_progress_done(loaded, self._zot.library_id)

        # TODO: Filter out the empty entries

        self.load_progress_done(loaded, self._zot.library_id)

        cache_file = self.get_cache_file_path()
        # Writing to file
        if cache_file:
            show_message(self._ls, f"Writing to bibfile to `{cache_file}`")
            bibtexparser.write_file(cache_file, self.library)

        return [self.library]



import logging
from dataclasses import dataclass, field

from attr import asdict

logger = logging.getLogger(__name__)


@dataclass
class Unionable:
    def __or__(self, other):
        return self.__class__(**asdict(self) | asdict(other))


"""Default header"""
DEFAULT_HEADER_FORMAT = [
    "# `{entry_type}` {title}",
    "_{author}_",
    "───────────────────────────────────────────────────────────────────────────────────",
]

"""Default footer"""
DEFAULT_FOOTER_FORMAT = [
    "───────────────────────────────────────────────────────────────────────────────────",
    "\n" + "\t" * 20 + "from `{bibfile}`",
]


"""Default cite trigger"""
DEFAULT_CITE_PRESET = "pandoc"

"""Default word wrap"""
DEFAULT_WRAP = 80

"""Default character limit"""
DEFAULT_CHAR_LIMIT = 400


@dataclass
class ViewConfig:
    """
    Configs for viewing documents
    """

    viewer: str = "browser"
    """`zotero`, `zotero_bbt` or `browser`"""


@dataclass
class DocFormatingConfig:
    """
    Configs for displaying documentation strings
    """

    wrap: int = DEFAULT_WRAP
    """Line wrap config"""

    character_limit: int = DEFAULT_CHAR_LIMIT
    """Number of characters before trimming"""

    format: str = "list"
    """`list` or `markdown`"""

    show_fields: list[str] = field(default_factory=lambda: [])
    """Filter of bibtex fields to show"""

    header_format: list[str] | str = field(
        default_factory=lambda: DEFAULT_HEADER_FORMAT
    )
    """
    List of Python-style format strings for the header.
    """

    footer_format: list[str] | str = field(
        default_factory=lambda: DEFAULT_FOOTER_FORMAT
    )
    """
    List of Python-style format strings for the footer.
    """


@dataclass
class CiteConfig(Unionable):
    """
    Configs for citation.
    """

    preset: str = "pandoc"
    """Trigger completion and also marks the beginning of citation key."""

    trigger: str = "@"
    """Trigger completion and also marks the beginning of citation key."""

    post_trigger: str = ","
    """Trigger completion and also marks the beginning of citation key."""

    prefix: str = "["
    """Prefix to begin a block of citation."""

    postfix: str = "]"
    """End of the citation block."""

    separator: str = ";"
    """separator between citations"""



@dataclass
class HoverConfig:
    """
    Configs for `textDocument/hover`.
    """

    doc_format: DocFormatingConfig = field(default_factory=lambda: DocFormatingConfig())
    """
    see DocFormatingConfig
    """


@dataclass
class CompletionConfig:
    """
    Configs for `textDocument/completion`.
    """

    doc_format: DocFormatingConfig = field(default_factory=lambda: DocFormatingConfig())


@dataclass
class BackendConfig:
    """
    Config for backends
    """

    backend_type: str = "bibfile"
    """Type of backend `bibfile` or `zotero_api`"""

    library_id: str = ""
    """`zotero_api` only: Online library ID"""

    library_type: str = "user"
    """`zotero_api` only: Online library type"""

    api_key: str | None = None
    """`zotero_api` only: API key """

    bibfiles: list[str] = field(default_factory=lambda: [])
    """`bibfile` only: List of bibfile paths to load"""




PANDOC_CITE_PRESET: CiteConfig = CiteConfig(
    preset="pandoc",
    trigger="@",
    post_trigger=",",
    prefix="[",
    postfix="]",
    separator=";",
)


CITE_PRESETS = {"pandoc": PANDOC_CITE_PRESET}


# TODO: Is there a better way to do this?
EXPECTED_VALUES = {
    "backend_type": ["zotero_api", "bibfile"],
    "library_type": ["user", "group"],
    "doc_format.format": ["table", "list"],
    "view.viewer": ["browser", "zotero", "zotero_bbt"],
}


@dataclass
class BibliTomlConfig:
    """
    All configurations used by bibli
    """

    backends: dict[str, BackendConfig] = field(default_factory=lambda: {})
    """Dictionary of backend configs"""

    hover: HoverConfig = field(default_factory=lambda: HoverConfig())
    """See `HoverConfig`"""

    completion: CompletionConfig = field(default_factory=lambda: CompletionConfig())
    """See `CompletionConfig`"""

    cite: CiteConfig = field(default_factory=lambda: CITE_PRESETS[DEFAULT_CITE_PRESET])
    """See `CiteConfig`"""

    view: ViewConfig = field(default_factory=lambda: ViewConfig())
    """See `ViewingConfig`"""

    def check_expected(self, field, value) -> bool:
        if value not in EXPECTED_VALUES[field]:
            logger.error(
                f"Unexpected value in {field}: {value}",
            )
            return False
        return True

    def sanitize(self) -> bool:
        valid = True
        for _, v in self.backends.items():
            valid |= self.check_expected("backend_type", v.backend_type)
            match v.backend_type:
                case "zotero_api":
                    valid |= self.check_expected("library_type", v.library_type)
                case "bibfile":
                    pass

        valid |= self.check_expected("doc_format.format", self.hover.doc_format.format)

        valid |= self.check_expected("view.viewer", self.view.viewer)

        valid |= self.check_expected(
            "doc_format.format", self.completion.doc_format.format
        )

        # Apply preset and user configs
        if self.cite.preset != DEFAULT_CITE_PRESET:
            if not CITE_PRESETS.get(self.cite.preset):
                logger.error(f"Unknown preset {self.cite.preset}")
                return False
            self.cite = CITE_PRESETS[self.cite.preset] | self.cite

        return valid
